﻿/*	File			NoughtsCrossesCLIENT.cs
	Purpose			Network Game (Noughts and Crosses)CLIENT
	Author			Richard John Anthony
	Date			January 2015 (original Cpp version December 2003)
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace CaseStudyGame_Client
{
    public partial class NoughtsCrossesCLIENT : Form
    {
        public const Int16 PORT = 8000;                     // Default port 8000 decimal
        public const Int16 SERVER_ADVERTISE_PORT = 8009;	// Server Advertisement port 8009 decimal
        public const Int32 MAX_ALIAS_LEN = 15;
        public const Int32 MAX_SERVER_NAME_LEN = 30;

        public enum MessageType {REGISTER_ALIAS, PLAYER_LIST, CHOOSE_OPPONENT, LOCAL_CELL_SELECTION,
        OPPONENT_CELL_SELECTION, START_OF_GAME, END_OF_GAME, CLIENT_DISCONNECT, SERVER_ADVERTISE};

        public enum CodeType {END_OF_GAME_DRAW, END_OF_GAME_WIN, END_OF_GAME_LOOSE};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Message_PDU
        {
	        public MessageType eMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_ALIAS_LEN + 1))]
	        public byte[] cAlias;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = ((MAX_ALIAS_LEN + 1)*10))]
	        public byte[] cPlayerList;      //comma-separated list of aliases
            public Int32 iCell;
            public CodeType eCode;
            public Int32 iToken;
            public bool bFirstMove;
        };
 
        // Service advertisement message structure - this is received from the Server, and needs to be deserialised
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct ServerAdvertise_Message_PDU
        {
	        public MessageType eMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_SERVER_NAME_LEN + 1))]
	        public byte[] cServerName;
	        public byte iIP_Address_Byte_1;	// LSB
            public byte iIP_Address_Byte_2;
            public byte iIP_Address_Byte_3;
            public byte iIP_Address_Byte_4;	// MSB
            // Note - Field alignments (compatibility with C++ version of the game) - Due to the nature of the serialisation
            // performed in the C++ version, all 4-byte fields are aligned on 4-byte-boundaries. The cServerName field is 
            // 31 bytes long, and thus ends not on a 4-byte boundary. The four single byte fields that follow it are aligned
            // on byte boundaries. This means that in the actual transmitted byte array, a single padding byte (value 0) is
            // inserted before the iPort field, this requires attention when serialising / deserialising in the C# versions 
            // of the application, see DeSerialize_ReceivedServerAdvertisementMessage()
            public Int32 iPort; // Port values are actually Int16, but encoded as Int32 in this message format for compatibility with the pre-existing C++ version
        };

        Socket m_ConnectSocket;
        Socket m_ListenForServerAdvertisementsSocket;
        IPEndPoint m_ConnectEndPoint;
        IPEndPoint m_ReceiveServerAdvertisementsEndPoint;
        System.Windows.Forms.Timer m_Receive_Timer;
        System.Windows.Forms.Timer m_ListenForserverAdvertisement_Timer;

        // Helper variables used in message Serialize, Deserialize and Receive
        Int32 m_iFieldSize1; // eMessageType
        Int32 m_iFieldSize2; // cAlias
        Int32 m_iFieldSize3; // cPlayerList
        Int32 m_iFieldSize4; // iCell
        Int32 m_iFieldSize5; // eCode
        Int32 m_iFieldSize6; // iToken
        Int32 m_iFieldSize7; // bFirstMove
        Int32 m_iPadSize_To4ByteBoundary; // Size of pad field to be added to meet a 4-byte boundary for interoperability with C++
        Int32 m_iMessageSize; // The structure itself, plus any padding

        // Helper variables used in Server Advertisement message Serialize, Deserialize and Receive
        Int32 m_iSA_FieldSize1; // eMessageType
        Int32 m_iSA_FieldSize2; // cServerName
        Int32 m_iSA_FieldSize3; // iIP_Address_Byte_1
        Int32 m_iSA_FieldSize4; // iIP_Address_Byte_2
        Int32 m_iSA_FieldSize5; // iIP_Address_Byte_3
        Int32 m_iSA_FieldSize6; // iIP_Address_Byte_4
        Int32 m_iSA_FieldSize7; // iPort
        Int32 m_iSA_PadSize_To4ByteBoundary; // Size of pad field to be added to meet a 4-byte boundary for interoperability with C++
        Int32 m_iSA_MessageSize; // The structure itself, plus any padding

	    public bool m_bConnected;
        string m_sToken;
	    string m_sOpponentToken;
  
        public NoughtsCrossesCLIENT()
        {
            InitializeComponent();
            m_bConnected = false;
            m_iMessageSize = CalculateMessageStructureSizeAndFieldSizes(ref m_iFieldSize1, ref m_iFieldSize2, ref m_iFieldSize3,
                        ref m_iFieldSize4, ref m_iFieldSize5, ref m_iFieldSize6, ref m_iFieldSize7, ref m_iPadSize_To4ByteBoundary);
            m_iSA_MessageSize = CalculateServerAdvertisementMessageStructureSizeAndFieldSizes(ref m_iSA_FieldSize1, ref m_iSA_FieldSize2, ref m_iSA_FieldSize3,
                        ref m_iSA_FieldSize4, ref m_iSA_FieldSize5, ref m_iSA_FieldSize6, ref m_iSA_FieldSize7, ref m_iSA_PadSize_To4ByteBoundary);
            InitialiseControls();
            Initialise_ListeningFor_ServerAdvertisement_broadcasts();
        }

        Int32 CalculateMessageStructureSizeAndFieldSizes(ref Int32 F1, ref Int32 F2, ref Int32 F3,
                             ref Int32 F4, ref Int32 F5, ref Int32 F6, ref Int32 F7, ref Int32 Pad)
        {  // Helper method, relating to the use of the Message_PDU message structure
            // simplifies message Serialize, Deserialize and Receive
            F1 = sizeof(MessageType);           // eMessageType
            F2 = MAX_ALIAS_LEN + 1;             // cAlias
            F3 = (MAX_ALIAS_LEN + 1) * 10;      // cPlayerList
            F4 = sizeof(Int32);                 // iCell
            F5 = sizeof(CodeType);              // eCode
            F6 = sizeof(Int32);                 // iToken
            F7 = sizeof(bool);                  // bFirstMove
            Int32 iContentMessageSize = F1 + F2 + F3 + F4 + F5 + F6 + F7;
            Pad = CalculatePadTo4ByteBoundary(iContentMessageSize);
            Int32 iPaddedMessageSize = iContentMessageSize + Pad;
            return iPaddedMessageSize;
        }

        Int32 CalculateServerAdvertisementMessageStructureSizeAndFieldSizes(ref Int32 F1, ref Int32 F2, ref Int32 F3,
                             ref Int32 F4, ref Int32 F5, ref Int32 F6, ref Int32 F7, ref Int32 Pad)
        {  // Helper method, relating to the use of the ServerAdvertise_Message_PDU message structure
            // simplifies message Serialize, Deserialize and Receive
            F1 = sizeof(MessageType);       // eMessageType
            F2 = MAX_SERVER_NAME_LEN + 1;   // cServerName
            F3 = 1;                         // byte iIP_Address_Byte_1;
            F4 = 1;                         // byte iIP_Address_Byte_2;
            F5 = 1;                         // byte iIP_Address_Byte_3;
            F6 = 1;                         // byte iIP_Address_Byte_4;
            F7 = sizeof(Int32);             // iPort, only needs Int16, but sent as Int32 (for compatibility with c++ versions)
            Int32 iContentMessageSize = F1 + F2 + F3 + F4 + F5 + F6 + F7;
            Pad = CalculatePadTo4ByteBoundary(iContentMessageSize);
            Int32 iPaddedMessageSize = iContentMessageSize + Pad;
            return iPaddedMessageSize;
        }

        Int32 CalculatePadTo4ByteBoundary(Int32 iContentMessageSize)
        {   // Helper method, relating to the use of the Message_PDU message structure
            // calculates the size of the padding neccesary to truncate the message on a 4-byte boundary (compatible with C++)
            // Pad will always be between 0 and 3 bytes
            int iRemainder = iContentMessageSize % 4; // The extent of overlap beyond 4-byte boundary
            if (0 == iRemainder)
            {   // Already on a 4-byte boundary
                return 0; // No padding necessary
            }
            else
            {   // Calculate pad to next 4-byte boundary
                return 4 - iRemainder;
            }
        }
           
        void InitialiseControls()
        {
            Port_textBox.Text = Convert.ToString(PORT);
            Port_textBox.Enabled = true;
            ConnectAddress_1_TextBox.Enabled = true;
            ConnectAddress_2_TextBox.Enabled = true;
            ConnectAddress_3_TextBox.Enabled = true;
            ConnectAddress_4_TextBox.Enabled = true;
            DisconnectButton.Enabled = false;
            RegisterAlias_Button.Enabled = false;
            AliasName_TextBox.Enabled = false;
            SelectPlayer_Button.Enabled = false;
            PlayerListBox.Enabled = false;
            GameStatus_TextBox.Text = "No active game"; // Display game status
            Token_TextBox.Text = "";
            MoveStatus_TextBox.Text = "";
            OpponentAlias_TextBox.Text = "";
            ConnectStatus_textBox.Text = "Waiting for Server";
            ClearAllCells();
        }

        IPAddress GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {
                    return ipHostAddress[iAddrIndex]; // Return the first IPv4 address found
                }
            }
            ConnectAddress_1_TextBox.Text = "No";
            ConnectAddress_2_TextBox.Text = "IP";
            ConnectAddress_3_TextBox.Text = "v4";
            ConnectAddress_4_TextBox.Text = "Addr";
            IPAddress EmptyIPAddress = new IPAddress(0);
            return EmptyIPAddress;
        }

        void CreateSocket()
        {
            try
            {
                m_ConnectSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ConnectSocket.Blocking = true; // Set blocking mode for socket during connection set-up phase
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "'Noughts and Crosses' Client");
                CloseSocketAndExit();
            }
        }

        void Connect_to_Server_using_received_AddressDetails()
        {	
            CreateSocket();

            try
            {
                m_ConnectSocket.Connect(m_ConnectEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Connect() failed - is server running?, are port and IP address values correct?", 
                                                                                "'Noughts and Crosses' Client");
                return;
            }
 
	        m_bConnected = true;
            Stop_ListenForserverAdvertisement_Timer();
            ConnectStatus_textBox.Text = "Connected to Server";

	        // Update controls
            Port_textBox.Enabled = false;
            ConnectAddress_1_TextBox.Enabled = false;
            ConnectAddress_2_TextBox.Enabled = false;
            ConnectAddress_3_TextBox.Enabled = false;
            ConnectAddress_4_TextBox.Enabled = false;
	        m_bConnected = true;
            DisconnectButton.Enabled = true;
	        AliasName_TextBox.Enabled = true;	
	        RegisterAlias_Button.Enabled = true;

	        // Set non-blocking (now that connection established) for receiving
            try
            {
                m_ConnectSocket.Blocking = false; // Set non-blocking mode
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Could not set non-Blocking behaviour for socket", "'Noughts and Crosses' Client");
            }

            StartTimer(); // Enable receiving (check periodically)
        }

        static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private byte[] Serialize_Message_PDU(Message_PDU Message)
        {   // Convert an Message_PDU structure into a byte array
            byte[] bByteArray = new byte[m_iMessageSize];        // Create a byte array to hold the serialised form of the message PDU

            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize1/*count*/);

            iFieldOffset += m_iFieldSize1;                  // Adjust offset for second field
            int iDataLength2 = Message.cAlias.Length;       // Get the length of data actually stored in field2 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cAlias/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength2/*count*/);

            iFieldOffset += m_iFieldSize2;                  // Adjust offset for third field
            int iDataLength3 = Message.cPlayerList.Length;  // Get the length of data actually stored in field3 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cPlayerList/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength3/*count*/);

            iFieldOffset += m_iFieldSize3;                  // Adjust offset for fourth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCell);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize4/*count*/);

            iFieldOffset += m_iFieldSize4;                  // Adjust offset for fifth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.eCode);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize5/*count*/);

            iFieldOffset += m_iFieldSize5;                  // Adjust offset for sixth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iToken);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize6/*count*/);

            iFieldOffset += m_iFieldSize6;                  // Adjust offset for seventh field
            bWorkingArray = BitConverter.GetBytes((bool)Message.bFirstMove); // Boolean is 1 byte
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize7/*count*/);

            // Encoding the padding (to ensure message size is aligned with a 4-byte boundary): The bByteArray has been created to the full message
            // size including the pad. This byte array is automatically initialised such that all bytes contain zeros, therefore no additional
            // encoding is necessary to include the padding at the end of the message. The padding is always between 0 and 3 bytes.
            return bByteArray;
        }

        bool SendMessage(Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectSocket.Send(bBuffer);
                if (0 >= iBytesSent)
                {
                    // Place diagnostics here if necessary
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {
                    // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   
                    if(false == Is_TCP_SocketConnected(m_ConnectSocket))
                    {
                    }
                    else // still connected
                    {
                        return false;
                    }
                }
            }
            return true;
         }
        
        void StartTimer()
        {
            m_Receive_Timer = new System.Windows.Forms.Timer();
            m_Receive_Timer.Interval = 100; // milliseconds
            m_Receive_Timer.Tick += OnTimedEvent_Receive_Timer;
            m_Receive_Timer.Enabled = true;
        }
 
        void Start_ListenForServerAdvertisement_Timer()
        {
            m_ListenForserverAdvertisement_Timer = new System.Windows.Forms.Timer();
            m_ListenForserverAdvertisement_Timer.Interval = 100; // milliseconds
            m_ListenForserverAdvertisement_Timer.Tick += OnTimedEvent_ListenForserverAdvertisement_Timer;
            m_ListenForserverAdvertisement_Timer.Enabled = true;
        }
        
        void Stop_Receive_Timer()
        {
            if (null != m_Receive_Timer)
            {
                m_Receive_Timer.Enabled = false;
            }
        }

        void Stop_ListenForserverAdvertisement_Timer()
        {
            if (null != m_ListenForserverAdvertisement_Timer)
            {
                m_ListenForserverAdvertisement_Timer.Enabled = false;
            }
        }

        void OnTimedEvent_Receive_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                DoReceive();
            }
            catch // Silently catch any errors
            {
            }
        }

        void OnTimedEvent_ListenForserverAdvertisement_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                DoReceive_ServerAdvertisement_broadcast();
            }
            catch // Silently catch any errors
            {
            }
        }

        void DoReceive()
        {
            byte[] bArray = new byte[m_iMessageSize];
            int iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount =  m_ConnectSocket.Receive(bArray);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   
                    if(false == Is_TCP_SocketConnected(m_ConnectSocket))
                    {
                    }
                    else // still connected
                    {
                        return;
                    }
                }
            }
 
 		    if(0 == iReceiveByteCount)
		    {
			    return;
		    }

            Message_PDU Message = DeSerialize_ReceivedMessage(bArray);

		    switch(Message.eMessageType)
		    {
                case MessageType.PLAYER_LIST:
		            DisplayPlayerList(Message);	
		            break;
                case MessageType.START_OF_GAME:
			        SelectPlayer_Button.Enabled = false;            // Disable Select Opponent control
			        string sOpponent = "";
                    int iLength = ByteArrayTo_ASCII_String(Message.cAlias, ref sOpponent);	// Display opponent alias
                    OpponentAlias_TextBox.Text = sOpponent;
                    
                    if(Message.iToken == 0) 						// Set and display our token
			        {
				        m_sToken = "0";
				        m_sOpponentToken = "X";
			        }
			        else
			        {
				        m_sToken = "X";
				        m_sOpponentToken = "0";
			        }		
			        Token_TextBox.Text = m_sToken;
			
			        if(Message.bFirstMove == true)					// configure first-move logic
			        {
				        MoveStatus_TextBox.Text = "Your Move";
				        EnableEmptyCells();
			        }
			        else
			        {
				        MoveStatus_TextBox.Text = "Opponent's Move";
				        DisableEmptyCells();
			        }
			        GameStatus_TextBox.Text = "Game In Progress"; // Display game status

 		            break;
                case MessageType.END_OF_GAME:
			        switch(Message.eCode) 
			        {
                        case CodeType.END_OF_GAME_DRAW:
			                MessageBox.Show("The game was a draw","Game Over");	
			                break;
                        case CodeType.END_OF_GAME_WIN:
                            MessageBox.Show("Well done! You won", "Game Over");	
			                break;
                        case CodeType.END_OF_GAME_LOOSE:
                            MessageBox.Show("Unlucky - You lost", "Game Over");	
			                break;
			        }
                    CloseSocketAndExit();
			        break;
                case MessageType.OPPONENT_CELL_SELECTION:
		            switch (Message.iCell)
			        {
		            case 0:
                        Cell0_Button.Text = m_sOpponentToken;
                        Cell0_Button.Enabled = false;
                        break;
		            case 1:
                        Cell1_Button.Text = m_sOpponentToken;
                        Cell1_Button.Enabled = false;
                        break;
		            case 2:
                        Cell2_Button.Text = m_sOpponentToken;
                        Cell2_Button.Enabled = false;
                        break;
		            case 3:
                        Cell3_Button.Text = m_sOpponentToken;
                        Cell3_Button.Enabled = false;
                        break;
		            case 4:
                        Cell4_Button.Text = m_sOpponentToken;
                        Cell4_Button.Enabled = false;
                        break;
		            case 5:
                        Cell5_Button.Text = m_sOpponentToken;
                        Cell5_Button.Enabled = false;
                        break;
		            case 6:
                        Cell6_Button.Text = m_sOpponentToken;
                        Cell6_Button.Enabled = false;
                        break;
		            case 7:
                        Cell7_Button.Text = m_sOpponentToken;
                        Cell7_Button.Enabled = false;
                        break;
		            case 8:
                        Cell8_Button.Text = m_sOpponentToken;
                        Cell8_Button.Enabled = false;
                        break;
			        }
			    MoveStatus_TextBox.Text = "Your Move";
			    EnableEmptyCells();
			    break;
                default:
                break;
		    }
	    }

        public Message_PDU DeSerialize_ReceivedMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
            Message_PDU result_Message = Message_PDU_Initialise();
            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);

            iFieldOffset += m_iFieldSize1;
            result_Message.cAlias = new byte[MAX_ALIAS_LEN + 1];
            for (int iIndex = 0; iIndex < m_iFieldSize2; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cAlias[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += m_iFieldSize2;
            result_Message.cPlayerList = new byte[(MAX_ALIAS_LEN + 1) * 10];
            for (int iIndex = 0; iIndex < m_iFieldSize3; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cPlayerList[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += m_iFieldSize3;
            result_Message.iCell = ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize4;
            result_Message.eCode = (CodeType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize5;
            result_Message.iToken = ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize6;
            result_Message.bFirstMove = BitConverter.ToBoolean(bRecdByteArray, iFieldOffset); // Converts a single byte within a byte array to a bool value   

            // Messages are natively aligned on 4-byte boundaries by C++ send(). To ensure compatibility, C# messages are adjusted in the
            // application code to truncate on 4-byte boundaries. The padding is added at the end of the message, so any bytes beyond the last
            // expected field can be ignored when receiving, i.e. no special decoding is needed for the padding when a message is DeSerializeed
            return result_Message;
        }

        void DisplayPlayerList(Message_PDU Message)
        {
	        PlayerListBox.Items.Clear();            // Clear contents of PlayerList control
	        string sAlias = AliasName_TextBox.Text; // Read our Alias (exclude it from displayed list)
            string sPlayerList = "";
            int ilength = ByteArrayTo_ASCII_String(Message.cPlayerList, ref sPlayerList);
            // sPlayerList is a comma-separated, ASCII (single byte per char) null-terminated list (terminated at iLength)
	        // There is no comma after the last item in the list
            int iNextCommaPos;
	        string sCurrentPlayer = "";
            while ((iNextCommaPos = sPlayerList.IndexOf(',')) > 0 && iNextCommaPos < ilength) // Ignore chars beyond the valid string content
            {
                sCurrentPlayer = sPlayerList.Substring(0/*start*/, iNextCommaPos/*length*/);
                if (!sCurrentPlayer.Equals(sAlias, StringComparison.OrdinalIgnoreCase)) // Exclude own alias from displayed list
                {
                    PlayerListBox.Items.Add(sCurrentPlayer);
                }
                    sPlayerList = sPlayerList.Substring(iNextCommaPos + 1/*start*/);// Remove 'dealt with' name from head of list
                    ilength -= (iNextCommaPos + 1);                                 // Keep track of remaining list-string length
	        }
            // Deal with residual substring (there is no comma after the last item in the list)
            sPlayerList = TruncateString(sPlayerList, ilength);
            if (0 < ilength && !sPlayerList.Equals(sAlias, StringComparison.OrdinalIgnoreCase)) // Exclude own alias from displayed list
            {
                 PlayerListBox.Items.Add(sPlayerList);
            }

            if (0 >= PlayerListBox.Items.Count)
            {
                PlayerListBox.Enabled = false;
                SelectPlayer_Button.Enabled = false;
            }
            else
            {
                PlayerListBox.SelectedIndex = 0; // Set default selection
                PlayerListBox.Enabled = true;
                SelectPlayer_Button.Enabled = true;
            }
        }

        void ConnectionLost()
        {
            CloseSocketAndExit();
        }

        private void DisconnectButton_Click(object sender, EventArgs e)
        {
	        if(m_bConnected)	// If connected - send disconnection message (graceful disconnect)
	        {
                Message_PDU Message = Message_PDU_Initialise(); 
                Message.eMessageType = MessageType.CLIENT_DISCONNECT;
		        SendMessage(Message);
	        }

	        CloseConnectionAndSocket();
	        InitialiseControls();
        }

        private void Done_Button_Click(object sender, EventArgs e)
        {
            CloseSocketAndExit();
        }

        private void RegisterAlias_Button_Click(object sender, EventArgs e) // Register Alias with server
        {
	        string sAlias = AliasName_TextBox.Text;
	        if(0 == sAlias.Length)
	        {
                MessageBox.Show("Alias name cannot be blank", "'Noughts and Crosses' Client");
	        }
	        else
	        {
                Message_PDU Message = Message_PDU_Initialise();
                Message.eMessageType = MessageType.REGISTER_ALIAS;
                sAlias = TruncateString(sAlias, MAX_ALIAS_LEN); // Truncate if necessary
                AliasName_TextBox.Text = sAlias; // Update textbox with (possibly) truncated string
                Message.cAlias = StringTo_ASCII_ByteArray(sAlias);
                if(SendMessage(Message) == true)
		        {
			        AliasName_TextBox.Enabled = false;	
			        RegisterAlias_Button.Enabled = false;
		        }
	        }
        }

        private void SelectPlayer_Button_Click(object sender, EventArgs e)
        {
            int iSelection = PlayerListBox.SelectedIndex;
	        if(-1 == iSelection)
	        {
                MessageBox.Show("No player selected", "'Noughts and Crosses' Client");
	        }
	        else
	        {
		        // Send a CHOOSE_OPPONENT message to server
                Message_PDU Message = Message_PDU_Initialise(); 
                Message.eMessageType = MessageType.CHOOSE_OPPONENT;
                Message.cAlias = StringTo_ASCII_ByteArray(PlayerListBox.SelectedItem.ToString());
                SendMessage(Message);
	        }
        }

       private void Cell0_Button_Click(object sender, EventArgs e)
        {
 	        Cell0_Button.Text = m_sToken;
	        Cell0_Button.Enabled = false;
	        CellSelectAction(0);
        }

        private void Cell1_Button_Click(object sender, EventArgs e)
        {
 	        Cell1_Button.Text = m_sToken;
	        Cell1_Button.Enabled = false;
	        CellSelectAction(1);
        }

        private void Cell2_Button_Click(object sender, EventArgs e)
        {
 	        Cell2_Button.Text = m_sToken;
	        Cell2_Button.Enabled = false;
	        CellSelectAction(2);
        }

        private void Cell3_Button_Click(object sender, EventArgs e)
        {
  	        Cell3_Button.Text = m_sToken;
	        Cell3_Button.Enabled = false;
	        CellSelectAction(3);
        }

        private void Cell4_Button_Click(object sender, EventArgs e)
        {
 	        Cell4_Button.Text = m_sToken;
	        Cell4_Button.Enabled = false;
	        CellSelectAction(4);
        }

        private void Cell5_Button_Click(object sender, EventArgs e)
        {
  	        Cell5_Button.Text = m_sToken;
	        Cell5_Button.Enabled = false;
	        CellSelectAction(5);
        }

        private void Cell6_Button_Click(object sender, EventArgs e)
        {
 	        Cell6_Button.Text = m_sToken;
	        Cell6_Button.Enabled = false;
	        CellSelectAction(6);
        }

        private void Cell7_Button_Click(object sender, EventArgs e)
        {
 	        Cell7_Button.Text = m_sToken;
	        Cell7_Button.Enabled = false;
	        CellSelectAction(7);
        }

        private void Cell8_Button_Click(object sender, EventArgs e)
        {
 	        Cell8_Button.Text = m_sToken;
	        Cell8_Button.Enabled = false;
	        CellSelectAction(8);
        }

        void CellSelectAction(int iCell)
        {   // send a CELL_SELECTION message to server
            Message_PDU Message = Message_PDU_Initialise(); 
            Message.eMessageType = MessageType.LOCAL_CELL_SELECTION;
	        Message.iCell = iCell;
            if(SendMessage(Message) == true)
	        {
                MoveStatus_TextBox.Text = "Opponent's Move";
		        DisableEmptyCells();
	        }
        }

        void DisableEmptyCells()
        {
	        if("" == Cell0_Button.Text) Cell0_Button.Enabled = false; 
	        if("" == Cell1_Button.Text) Cell1_Button.Enabled = false; 
	        if("" == Cell2_Button.Text) Cell2_Button.Enabled = false; 
	        if("" == Cell3_Button.Text) Cell3_Button.Enabled = false; 
	        if("" == Cell4_Button.Text) Cell4_Button.Enabled = false; 
	        if("" == Cell5_Button.Text) Cell5_Button.Enabled = false; 
	        if("" == Cell6_Button.Text) Cell6_Button.Enabled = false; 
	        if("" == Cell7_Button.Text) Cell7_Button.Enabled = false; 
	        if("" == Cell8_Button.Text) Cell8_Button.Enabled = false; 
        }

        void EnableEmptyCells()
        {
	        if("" == Cell0_Button.Text) Cell0_Button.Enabled = true; 
	        if("" == Cell1_Button.Text) Cell1_Button.Enabled = true; 
	        if("" == Cell2_Button.Text) Cell2_Button.Enabled = true; 
	        if("" == Cell3_Button.Text) Cell3_Button.Enabled = true; 
	        if("" == Cell4_Button.Text) Cell4_Button.Enabled = true; 
	        if("" == Cell5_Button.Text) Cell5_Button.Enabled = true; 
	        if("" == Cell6_Button.Text) Cell6_Button.Enabled = true; 
	        if("" == Cell7_Button.Text) Cell7_Button.Enabled = true; 
	        if("" == Cell8_Button.Text) Cell8_Button.Enabled = true; 
        }

        void ClearAllCells()
        {
            Cell0_Button.Text = "";
            Cell1_Button.Text = "";
            Cell2_Button.Text = "";
            Cell3_Button.Text = "";
            Cell4_Button.Text = "";
            Cell5_Button.Text = "";
            Cell6_Button.Text = "";
            Cell7_Button.Text = "";
            Cell8_Button.Text = "";
	        DisableEmptyCells();
        }        
        
        void CloseSocketAndExit()
        {
            CloseConnectionAndSocket();
            Application.Exit();
        }

        void CloseConnectionAndSocket()
        {
            Stop_Receive_Timer();
            Stop_ListenForserverAdvertisement_Timer();

            try
            {
                m_ConnectSocket.Shutdown(SocketShutdown.Both);	// Close the connection (in both directions)
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
            try
            {
                m_ConnectSocket.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
            try
            {
                m_ListenForServerAdvertisementsSocket.Close();
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        Message_PDU Message_PDU_Initialise()
        {
            Message_PDU M = new Message_PDU();
            M.eMessageType = MessageType.END_OF_GAME; // This field is always overwritten as used in every instance
            M.cAlias = StringTo_ASCII_ByteArray("");
            M.cPlayerList = StringTo_ASCII_ByteArray("");
            M.iCell = 0;
            M.eCode = CodeType.END_OF_GAME_DRAW;
            M.iToken = 0;
            M.bFirstMove = false;
            return M;
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            //Encoding.ASCII.GetBytes(sStr);
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }

        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if (sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }

        void InitialiseLocalAddressStructAndPort()
        {	// Initialise Local IP address and Port number for listening for Server Advertisement messages
            IPAddress ReceiveServerAdvertisementsAddress = GetLocalHostAddressDetails();
            m_ReceiveServerAdvertisementsEndPoint = new IPEndPoint(ReceiveServerAdvertisementsAddress, SERVER_ADVERTISE_PORT);
        }

        void Initialise_ListeningFor_ServerAdvertisement_broadcasts()
        {
	        // Create Listen Socket and set it NonBlocking
            try
            {
                m_ListenForServerAdvertisementsSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_ListenForServerAdvertisementsSocket.Blocking = false;
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "'Noughts and Crosses' Client");
                CloseSocketAndExit();
            }
	
	        // Set up local address structure
	        InitialiseLocalAddressStructAndPort(); // For listening for Server Advertisement broadcasts

	        // Bind   *** NOTE THAT BECAUSE THE CLIENT BINDS TO A SPECIFIC PORT, THIS SOLUTION ONLY PERMITS ONE CLIENT INSTANCE TO RUN AT EACH COMPUTER ***
            try
            {
	            m_ListenForServerAdvertisementsSocket.Bind(m_ReceiveServerAdvertisementsEndPoint);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("Bind failed !\nPort is probably in use", "'Noughts and Crosses' Client");
                CloseSocketAndExit();
            }

            Start_ListenForServerAdvertisement_Timer(); // Enable ServerAdvertisement_Reception (see timer logic)
        }

        void DoReceive_ServerAdvertisement_broadcast() // Invoked by timer
        {	
            byte[] bArray = new byte[m_iSA_MessageSize];
            int iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount =  m_ListenForServerAdvertisementsSocket.Receive(bArray);
            }
            catch (SocketException)
            {
                return;
            }
 
 		    if(0 == iReceiveByteCount)
		    {
			    return;
		    }

            ServerAdvertise_Message_PDU Message = DeSerialize_ReceivedServerAdvertisementMessage(bArray);

 		    if(MessageType.SERVER_ADVERTISE == Message.eMessageType)
		    {
                ConnectStatus_textBox.Text = "Server discovered";
                string sServerName = "";
                int ilength = ByteArrayTo_ASCII_String(Message.cServerName, ref sServerName);
                ServerName_textBox.Text = sServerName;

                // Set Server's Address
                byte[] bAddressArray = new byte[4];
                bAddressArray[0] = Message.iIP_Address_Byte_1;
                bAddressArray[1] = Message.iIP_Address_Byte_2;
                bAddressArray[2] = Message.iIP_Address_Byte_3;
                bAddressArray[3] = Message.iIP_Address_Byte_4;
                Int16 iPort = (Int16) Convert.ToInt32(Message.iPort);
                IPAddress ipHostAddress = new IPAddress(bAddressArray);
                m_ConnectEndPoint = new System.Net.IPEndPoint(ipHostAddress, iPort);

                // Display Server's Address
                ConnectAddress_1_TextBox.Text = Message.iIP_Address_Byte_1.ToString();
                ConnectAddress_2_TextBox.Text = Message.iIP_Address_Byte_2.ToString();
                ConnectAddress_3_TextBox.Text = Message.iIP_Address_Byte_3.ToString();
                ConnectAddress_4_TextBox.Text = Message.iIP_Address_Byte_4.ToString();
                Port_textBox.Text = Message.iPort.ToString();

			    Connect_to_Server_using_received_AddressDetails();
		    }
	    }

        public ServerAdvertise_Message_PDU DeSerialize_ReceivedServerAdvertisementMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
            ServerAdvertise_Message_PDU result_Message = new ServerAdvertise_Message_PDU();    
            
            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);

            iFieldOffset += m_iSA_FieldSize1;
            result_Message.cServerName = new byte[MAX_SERVER_NAME_LEN + 1];
            for (int iIndex = 0; iIndex < m_iSA_FieldSize2; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cServerName[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += (MAX_SERVER_NAME_LEN + 1);
            result_Message.iIP_Address_Byte_1 = bRecdByteArray[iFieldOffset++];
            result_Message.iIP_Address_Byte_2 = bRecdByteArray[iFieldOffset++];
            result_Message.iIP_Address_Byte_3 = bRecdByteArray[iFieldOffset++];
            result_Message.iIP_Address_Byte_4 = bRecdByteArray[iFieldOffset++];
            
            // The deserialisation has been designed to be compatible with the C++ version of the application. The C++ serialisation is based on a
            // direct flattening of a structure containing fixed-size fields. The four individual byte values that hold the address are not aligned
            // on 4-byte boundaries, but the following port number IS. Therefore, a byte of padding is inserted in the message (at the sender side) 
            // to push the port field onto the next 4-byte boundary.
            iFieldOffset++; // Skip the single byte of padding

            // Deserialisation of port number
            // The message format and deserialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            result_Message.iPort = (256 * bRecdByteArray[iFieldOffset])/*1st of the 4 bytes*/ + bRecdByteArray[iFieldOffset + 1]/*2nd of the 4 bytes*/;
            return result_Message;
        }
    }
}
