﻿/*	File			NoughtsCrossesSERVER.cs
	Purpose			Network Game (Noughts and Crosses)SERVER
	Author			Richard John Anthony
	Date			January 2015 (original Cpp version December 2003)
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace CaseStudyGame_Server
{
    public partial class GameServerDialog : Form
    {
        public const Int16 PORT = 8000;                     // Default port 8000 decimal
        public const Int16 SERVER_ADVERTISE_PORT = 8009;	// Server Advertisement port 8009 decimal
        public const Int32 MAX_ALIAS_LEN = 15;
        public const Int32 MAX_SERVER_NAME_LEN = 30;

        public enum MessageType {REGISTER_ALIAS, PLAYER_LIST, CHOOSE_OPPONENT, LOCAL_CELL_SELECTION,
                OPPONENT_CELL_SELECTION, START_OF_GAME, END_OF_GAME, CLIENT_DISCONNECT, SERVER_ADVERTISE};

        public enum CodeType {END_OF_GAME_DRAW, END_OF_GAME_WIN, END_OF_GAME_LOSE, END_OF_GAME_CLIENT_DISCONNECT};

        // Message structure (this will be serialised before sending)
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct Message_PDU
        {
	        public MessageType eMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_ALIAS_LEN + 1))]
	        public byte[] cAlias;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = ((MAX_ALIAS_LEN + 1)*10))]
	        public byte[] cPlayerList;      // Comma-separated list of aliases
            public Int32 iCell;
            public CodeType eCode;
            public Int32 iToken;
            public bool bFirstMove;
        };
        
        // Service advertisement message structure - this is received from the Server, and needs to be deserialised
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public unsafe struct ServerAdvertise_Message_PDU
        {
	        public MessageType eMessageType;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = (MAX_SERVER_NAME_LEN + 1))]
	        public byte[] cServerName;
	        public byte iIP_Address_Byte_1;	// LSB
            public byte iIP_Address_Byte_2;
            public byte iIP_Address_Byte_3;
            public byte iIP_Address_Byte_4;	// MSB
            // Note - Field alignments (compatibility with C++ version of the game) - Due to the nature of the serialisation
            // performed in the C++ version, all 4-byte fields are aligned on 4-byte-boundaries. The cServerName field is 
            // 31 bytes long, and thus ends not on a 4-byte boundary. The four single byte fields that follow it are aligned
            // on byte boundaries. This means that in the actual transmitted byte array, a single padding byte (value 0) is
            // inserted before the iPort field, this requires attention when serialising / deserialising in the C# versions 
            // of the application, see DeSerialize_ReceivedServerAdvertisementMessage()
            public Int32 iPort; // Port values are actually Int16, but encoded as Int32 in this message format for compatibility with the pre-existing C++ version
        };
                
        Socket m_ListenSocket;
        Socket m_SendServerAdvertisementsSocket;
        Int16 m_iPort;
        IPEndPoint m_LocalEndPoint;
        IPEndPoint m_SendServerAdvertisementsEndPoint;
        System.Windows.Forms.Timer m_Receive_Timer;
        System.Windows.Forms.Timer m_SendServerAdvertisement_Timer;

        // Helper variables used in message Serialize, Deserialize and Receive
        Int32 m_iFieldSize1; // eMessageType
        Int32 m_iFieldSize2; // cAlias
        Int32 m_iFieldSize3; // cPlayerList
        Int32 m_iFieldSize4; // iCell
        Int32 m_iFieldSize5; // eCode
        Int32 m_iFieldSize6; // iToken
        Int32 m_iFieldSize7; // bFirstMove
        Int32 m_iPadSize_To4ByteBoundary; // Size of pad field to be added to meet a 4-byte boundary for interoperability with C++
        Int32 m_iMessageSize; // The structure itself, plus any padding

        // Helper variables used in Server Advertisement message Serialize, Deserialize and Receive
        Int32 m_iSA_FieldSize1; // eMessageType
        Int32 m_iSA_FieldSize2; // cServerName
        Int32 m_iSA_FieldSize3; // iIP_Address_Byte_1
        Int32 m_iSA_FieldSize4; // iIP_Address_Byte_2
        Int32 m_iSA_FieldSize5; // iIP_Address_Byte_3
        Int32 m_iSA_FieldSize6; // iIP_Address_Byte_4
        Int32 m_iSA_FieldSize7; // iPort
        Int32 m_iSA_PadSize_To4ByteBoundary; // Size of pad field to be added to meet a 4-byte boundary for interoperability with C++
        Int32 m_iSA_MessageSize; // The structure itself, plus any padding

        const Int16 MAX_GAMES = 5;
        const Int16 MAX_CONNECTIONS = MAX_GAMES * 2;

        struct Connection {
	        public Socket ConnectSock;
	        public string sAlias;
	        public bool bInUse;
        };

        struct Game {
	        public int[] iConnection;   // Two connections (two clients) per game
	        public bool bInUse;
	        public int[,] iGrid;        // 3 x 3 noughts and crosses game grid
        };
        
	    Connection[] m_ConnectionArray;
	    Game[] m_GameArray;

        Random m_Random; // Needed for determining which player has first turn in a newly-created game

        public GameServerDialog()
        {
            InitializeComponent();
            m_iMessageSize = CalculateMessageStructureSizeAndFieldSizes(ref m_iFieldSize1, ref m_iFieldSize2,
                    ref m_iFieldSize3, ref m_iFieldSize4, ref m_iFieldSize5, ref m_iFieldSize6, ref m_iFieldSize7, ref m_iPadSize_To4ByteBoundary);
            m_iSA_MessageSize = CalculateServerAdvertisementMessageStructureSizeAndFieldSizes(ref m_iSA_FieldSize1, ref m_iSA_FieldSize2,
                    ref m_iSA_FieldSize3, ref m_iSA_FieldSize4, ref m_iSA_FieldSize5, ref m_iSA_FieldSize6, ref m_iSA_FieldSize7, ref m_iSA_PadSize_To4ByteBoundary); 

            m_ConnectionArray = new Connection[MAX_CONNECTIONS];
	        m_GameArray = new Game [MAX_GAMES];
 	        m_Random = new Random(DateTime.Now.Millisecond/*Seed*/); //Seed Random Number Generator;

	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
                m_ConnectionArray[i].bInUse = false;
                m_ConnectionArray[i].sAlias = "";
 	        }	
	        for(int i = 0; i < MAX_GAMES; i++)
	        {	
		        m_GameArray[i].iConnection = new int[2];
                m_GameArray[i].bInUse = false;
                m_GameArray[i].iGrid = new int[3,3];
	        }
	        InitialiseControls();
       }

        Int32 CalculateMessageStructureSizeAndFieldSizes(ref Int32  F1, ref Int32 F2, ref Int32 F3,
                            ref Int32 F4, ref Int32 F5, ref Int32 F6, ref Int32 F7, ref Int32 Pad) 
        {  // Helper method, relating to the use of the Message_PDU message structure
            // simplifies message Serialize, Deserialize and Receive
            F1 = sizeof(MessageType);           // eMessageType
            F2 = MAX_ALIAS_LEN + 1;             // cAlias
            F3 = (MAX_ALIAS_LEN + 1) * 10;      // cPlayerList
            F4 = sizeof(Int32);                 // iCell
            F5 = sizeof(CodeType);              // eCode
            F6 = sizeof(Int32);                 // iToken
            F7 = sizeof(bool);                  // bFirstMove
            Int32 iContentMessageSize = F1 + F2 + F3 + F4 + F5 + F6 + F7;
            Pad = CalculatePadTo4ByteBoundary(iContentMessageSize);
            Int32 iPaddedMessageSize = iContentMessageSize + Pad;
            return iPaddedMessageSize;
        }

        Int32 CalculateServerAdvertisementMessageStructureSizeAndFieldSizes(ref Int32 F1, ref Int32 F2, ref Int32 F3,
                             ref Int32 F4, ref Int32 F5, ref Int32 F6, ref Int32 F7, ref Int32 Pad)
        {  // Helper method, relating to the use of the ServerAdvertise_Message_PDU message structure
            // simplifies message Serialize, Deserialize and Receive
            F1 = sizeof(MessageType);       // eMessageType
            F2 = MAX_SERVER_NAME_LEN + 1;   // cServerName
            F3 = 1;                         // byte iIP_Address_Byte_1;
            F4 = 1;                         // byte iIP_Address_Byte_2;
            F5 = 1;                         // byte iIP_Address_Byte_3;
            F6 = 1;                         // byte iIP_Address_Byte_4;
            F7 = sizeof(Int32);             // iPort, only needs Int16, but sent as Int32 (for compatibility with c++ versions)
            Int32 iContentMessageSize = F1 + F2 + F3 + F4 + F5 + F6 + F7;
            Pad = CalculatePadTo4ByteBoundary(iContentMessageSize);
            Int32 iPaddedMessageSize = iContentMessageSize + Pad;
            return iPaddedMessageSize;
        }
        
        Int32 CalculatePadTo4ByteBoundary(Int32 iContentMessageSize)
        {   // Helper method, relating to the use of the Message_PDU message structure
            // calculates the size of the padding neccesary to truncate the message on a 4-byte boundary (compatible with C++)
            // Pad will always be between 0 and 3 bytes
            int iRemainder = iContentMessageSize % 4; // The extent of overlap beyond 4-byte boundary
            if (0 == iRemainder)
            {   // Already on a 4-byte boundary
                return 0; // No padding necessary
            }
            else
            {   // Calculate pad to next 4-byte boundary
                return 4 - iRemainder;
            }
        }
 
        void InitialiseControls()
        {
            Port_TextBox.Text = Convert.ToString(PORT);
            Port_TextBox.Enabled = true;
	        IP_Address_TextBox.Enabled = false;
            EnableServer_Button.Enabled = true;
	        CountActiveConnections();   // Updates control
	        CountActiveGames();         // Updates control
        }

        IPAddress GetLocalHostAddressDetails()
        {
            string sHostName = Dns.GetHostName();                   // Get the local host's name
            IPHostEntry ipHostEntry = Dns.GetHostEntry(sHostName);  // Get the host's address details
            IPAddress[] ipHostAddress = ipHostEntry.AddressList;    // Extract the IP address
            // ipHostAddress is an array of multiple addresses, we take the first IPv4 address from this list
            for (int iAddrIndex = 0; iAddrIndex < ipHostAddress.Length; iAddrIndex++)
            {   // Iterate through list of addresses, looking for the first IPv4 address
                if (ipHostAddress[iAddrIndex].AddressFamily == AddressFamily.InterNetwork)
                {
                    byte[] bAddressBytes = ipHostAddress[iAddrIndex].GetAddressBytes();
                    IP_Address_TextBox.Text = ipHostAddress[iAddrIndex].ToString();
                    return ipHostAddress[iAddrIndex];
                }
            }
            // No IPv4 address found - local host will not be able to connect to server using IPv4
            IP_Address_TextBox.Text = "No IPv4 Addr";
            EnableServer_Button.Enabled = false;
            return null;    
        }

        void CreateSocket()
        {
            try
            {
                m_ListenSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                m_ListenSocket.Blocking = false; // Non-blocking mode
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "'Noughts and Crosses' Server");
                CloseAll();
            }
        }

        void ReadFinalPortValue()
        {	
	        // The Port number can be overwritten by the user since it was
	        // automatically set to the default value. Hence the need to read the final value
	        m_iPort = Convert.ToInt16(Port_TextBox.Text);
        }

        void Accept()
        {
	        int iIndex = GetIndexOfUnusedConnection();
	        if(iIndex == -1)
	        {
		        WriteStatusLine("Accept refused - Maximum number of Connections reached");
		        return;
	        }
	
	        try
            {
	            m_ConnectionArray[iIndex].ConnectSock = m_ListenSocket.Accept();
	        }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   
			        MessageBox.Show("Accept failed","CNoughtsCrossesSERVER");
		        }
                return;
	        }
	        try
            {   // Set the newly-created connection socket to Non-Blocking
		        m_ConnectionArray[iIndex].ConnectSock.Blocking = false;
		    }
            catch(SocketException)
            {
			        MessageBox.Show("Could not set non-Blocking behaviour for connection socket","CNoughtsCrossesSERVER");
			        return;	
		    }

		    m_ConnectionArray[iIndex].bInUse = true;
		    string sStr = string.Format("Client {0} Connected", m_ConnectionArray[iIndex].ConnectSock.RemoteEndPoint.ToString());
		    WriteStatusLine(sStr);
		    CountActiveConnections(); // Updates control
        }

        private void EnableServer_Button_Click(object sender, EventArgs e)
        {
	        // Create Listen Socket and set it NonBlocking
            CreateSocket();
	        
            // Read the port value because the user may have changed it from the default
	        ReadFinalPortValue();

            // Set local endPoint Address and Port
            IPAddress LocalIPAddress = GetLocalHostAddressDetails();
            if (null == LocalIPAddress)
            {
                MessageBox.Show("Could not get local IP address", "CNoughtsCrossesSERVER");
                return;
            }
            m_LocalEndPoint = new IPEndPoint(LocalIPAddress, m_iPort);

	        // Bind
	        try
            {
                m_ListenSocket.Bind(m_LocalEndPoint);
            }
            catch
            {
		        MessageBox.Show("Bind failed !\nIs Port already in use?","CNoughtsCrossesSERVER");
		        return;
	        }

	        // Listen - for TCP connection requests from clients
	        try
            {
                m_ListenSocket.Listen(5/*Backlog*/);
            }
            catch
            {
		        MessageBox.Show("Listen failed","CNoughtsCrossesSERVER");
		        return;
	        }
		    EnableServer_Button.Enabled = false;
            Initialise_ServerAdvertisement();
            Start_Receive_Timer();   // All activity is non-blocking
        }

        int GetIndexOfUnusedConnection()
        {
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        if(m_ConnectionArray[i].bInUse == false)
		        {
			        return i;
		        }
	        }
	        return -1;	// Signal that all connections are in use
        }

        int CountActiveConnections()
        {
	        int iCount = 0;
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        if(m_ConnectionArray[i].bInUse == true)
		        {
			        iCount++;;
		        }
	        }
            NumConnectedClients_TextBox.Text = iCount.ToString(); // update number of connections control
 	        return iCount;
        }

        int CountActiveGames()
        {
	        int iCount = 0;
	        for(int i = 0; i < MAX_GAMES; i++)
	        {	
		        if(m_GameArray[i].bInUse == true)
		        {
			        iCount++;;
		        }
	        }
            this.NumActiveGames_TextBox.Text = iCount.ToString(); // update number of games control
	        return iCount;
        }

        int FindIndexOfConnectionForAlias(string sAlias)
        {
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        if(true == m_ConnectionArray[i].sAlias.Equals(sAlias,StringComparison.OrdinalIgnoreCase))
		        {
			        return i;
		        }
	        }
	        return -1;	// Signal that Alias name was not found in connection array
        }

        int GetIndexOfUnusedGame()
        {
	        for(int i = 0; i < MAX_GAMES; i++)
	        {	
		        if(m_GameArray[i].bInUse == false)
		        {
			        return i;
		        }
	        }
	        return -1;	// Signal that all connections are in use
        }

        bool IsConnectionPartOfActiveGame(int iConnection)
        {
	        for(int i = 0; i < MAX_GAMES; i++)
	        {	
		        if(m_GameArray[i].bInUse == true)
		        {
			        if(m_GameArray[i].iConnection[0] == iConnection	||
				        m_GameArray[i].iConnection[1] == iConnection)
			        {	
				        return true;
			        }
		        }
	        }
	        return false;	// Signal connections is not part of an active game
        }

        int GetConnectionOpponent(int iConnection)
        {
	        for(int i = 0; i < MAX_GAMES; i++)
	        {	
		        if(m_GameArray[i].bInUse == true)
		        {
			        if(m_GameArray[i].iConnection[0] == iConnection)
			        {
				        return m_GameArray[i].iConnection[1];
			        }	
			        if(m_GameArray[i].iConnection[1] == iConnection)
			        {
				        return m_GameArray[i].iConnection[0];
			        }
		        }
	        }
            return -1;
        }

        void Start_Receive_Timer()
        {
            m_Receive_Timer = new System.Windows.Forms.Timer();
            m_Receive_Timer.Interval = 100; // milliseconds (Short interval to allow UI responsiveness)
            m_Receive_Timer.Tick += OnTimedEvent_Receive_Timer;
            m_Receive_Timer.Enabled = true;
        }

        void Start_SendServerAdvertisement_Timer()
        {
            m_SendServerAdvertisement_Timer = new System.Windows.Forms.Timer();
            m_SendServerAdvertisement_Timer.Interval = 1000; // 1 second interval
            m_SendServerAdvertisement_Timer.Tick += OnTimedEvent_SendServerAdvertisement_Timer;
            m_SendServerAdvertisement_Timer.Enabled = true;
        }

        void StopTimer()
        {
            if (null != m_Receive_Timer)
            {
                m_Receive_Timer.Enabled = false;
            }
        }

        void OnTimedEvent_Receive_Timer(Object myObject, EventArgs myEventArgs)
        {
            try
            {
                Accept();
                for (int i = 0; i < MAX_CONNECTIONS; i++)
                {
                    if (m_ConnectionArray[i].bInUse == true)
                    {
                        DoReceive(i);	// For each valid listen-socket DoReceive()
                    }
                }
            }
            catch // Silently catch any errors
            {
            }
        }

        void OnTimedEvent_SendServerAdvertisement_Timer(Object myObject, EventArgs myEventArgs)
        {
            Send_ServerAdvertisement_Broadcast();
        }

        bool SendMessage(int iConnection, Message_PDU Message)
        {
            byte[] bBuffer = Serialize_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_ConnectionArray[iConnection].ConnectSock.Send(bBuffer);
                if (0 < iBytesSent)
                {
                }
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   // check if connection has been lost  
                    if (ex.NativeErrorCode.Equals(10053/*WSAECONNABORTED*/) ||
                        ex.NativeErrorCode.Equals(10054/*WSAECONNRESET*/) ||
                        ex.NativeErrorCode.Equals(10060/*WSAETIMEDOUT*/) ||
                        ex.NativeErrorCode.Equals(10052/*WSAENETRESET*/) ||
                        ex.NativeErrorCode.Equals(10058/*WSAESHUTDOWN*/))
                    {
                        CloseGameAndConnection(iConnection);
                    }
                }
            }
            string sStr = string.Format("Sent: {0} to:{1}", Message.eMessageType, m_ConnectionArray[iConnection].ConnectSock.RemoteEndPoint.ToString());
            WriteStatusLine(sStr);
            return true;
         }

	    static bool Is_TCP_SocketConnected(Socket sSock)
        {
            if (sSock.ProtocolType != ProtocolType.Tcp)
            {   // This test is necessary because the Connected property may return true for a UDP
                // socket (which is meaningless as it is alsways 'connectionless')
                return false;
            }
            // Note, the Connected property alone is not reliable. It indicates the connected state of the socket
            // at the time of executing the most recent Send or Receive operation
            return !((sSock.Poll(1000, SelectMode.SelectRead) && (sSock.Available == 0)) || !sSock.Connected);
        }

        private byte[] Serialize_Message_PDU(Message_PDU Message)
        {   // Convert an Message_PDU structure into a byte array
            byte[] bByteArray = new byte[m_iMessageSize];        // Create a byte array to hold the serialised form of the message PDU

            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize1/*count*/);

            iFieldOffset += m_iFieldSize1;                  // Adjust offset for second field
            int iDataLength2 = Message.cAlias.Length;       // Get the length of data actually stored in field2 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cAlias/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength2/*count*/);

            iFieldOffset += m_iFieldSize2;                  // Adjust offset for third field
            int iDataLength3 = Message.cPlayerList.Length;  // Get the length of data actually stored in field3 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cPlayerList/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength3/*count*/);

            iFieldOffset += m_iFieldSize3;                  // Adjust offset for fourth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iCell);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize4/*count*/);
 
            iFieldOffset += m_iFieldSize4;                  // Adjust offset for fifth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.eCode);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize5/*count*/);

            iFieldOffset += m_iFieldSize5;                  // Adjust offset for sixth field
            bWorkingArray = BitConverter.GetBytes((Int32)Message.iToken);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize6/*count*/);

            iFieldOffset += m_iFieldSize6;                  // Adjust offset for seventh field
            bWorkingArray = BitConverter.GetBytes((bool)Message.bFirstMove); // Boolean is 1 byte
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iFieldSize7/*count*/);

            // Encoding the padding (to ensure message size is aligned with a 4-byte boundary): The bByteArray has been created to the full message
            // size including the pad. This byte array is automatically initialised such that all bytes contain zeros, therefore no additional
            // encoding is necessary to include the padding at the end of the message. The padding is always between 0 and 3 bytes.
            return bByteArray;
        }
        
        void SendPlayerList()
        {
	        Message_PDU Message = Message_PDU_Initialise(); 
	        Message.eMessageType = MessageType.PLAYER_LIST;
            string sPlayerList = "";
	        // List should include all Aliases that are connected, but not involved in active games
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        if(m_ConnectionArray[i].bInUse == true && IsConnectionPartOfActiveGame(i) == false &&
			        m_ConnectionArray[i].sAlias.Equals("") == false)
		        {
                    sPlayerList += m_ConnectionArray[i].sAlias;
			        sPlayerList += ","; // Needs to be a comma-separated list
		        }
	        }
	        if(sPlayerList.Length > 0)
	        {
                sPlayerList = TruncateString(sPlayerList, sPlayerList.Length - 1); // Null terminate (write over last comma)
	        }

            Message.cPlayerList = StringTo_ASCII_ByteArray(sPlayerList);
            // Send PlayerList to ALL connected players
	        for(int i = 0; i < MAX_CONNECTIONS; i++)
	        {	
		        if(m_ConnectionArray[i].bInUse == true)
		        {
                    SendMessage(i, Message);
		        }
	        }
        }

        void SendOpponentCellSelectionMessage(int iConnection, int iCell)
        {
	        int iOpponentConnection;
	        iOpponentConnection = GetConnectionOpponent(iConnection);	// Find players opponent
	        if(iOpponentConnection < 0)
	        {
		        string sStr = string.Format("Opponent not found for connection {0}",iConnection);
		        WriteStatusLine(sStr);
		        return;
	        }
	        Message_PDU Message = Message_PDU_Initialise(); 
	        Message.eMessageType = MessageType.OPPONENT_CELL_SELECTION;
	        Message.iCell = iCell;
            SendMessage(iOpponentConnection, Message);
        }

        void DoReceive(int iConnection)
        {		
            byte[] bArray = new byte[m_iMessageSize];
            int iReceiveByteCount = 0;
            try
            {
                iReceiveByteCount =  m_ConnectionArray[iConnection].ConnectSock.Receive(bArray);
            }
            catch (SocketException ex)
            {
                if (ex.NativeErrorCode.Equals(10035/*WSAEWOULDBLOCK*/))
                {   // IGNORE - The exception was 'WOULD BLOCK' 
                }
                else
                {   // check if connection has been lost  
        	        if( ex.NativeErrorCode.Equals(10053/*WSAECONNABORTED*/) ||
                        ex.NativeErrorCode.Equals(10054/*WSAECONNRESET*/) ||
                        ex.NativeErrorCode.Equals(10060/*WSAETIMEDOUT*/) ||
                        ex.NativeErrorCode.Equals(10052/*WSAENETRESET*/) ||
                        ex.NativeErrorCode.Equals(10058/*WSAESHUTDOWN*/))                       
	                {
		                CloseGameAndConnection(iConnection);
	                }
                }
            }              

 		    if(0 == iReceiveByteCount)
		    {
			    return;
		    }

            Message_PDU Message = DeSerialize_ReceivedMessage(bArray);

		    string sStr = "";
            string sAlias = "";
            int ilength;
            switch (Message.eMessageType)
		    {
		        case MessageType.REGISTER_ALIAS:
                    ilength = ByteArrayTo_ASCII_String(Message.cAlias, ref sAlias);
                    if (0 < ilength)
                    {
                        sAlias = TruncateString(sAlias, ilength);
                        sStr = string.Format("Alias '{0}' Registered", sAlias);
                        WriteStatusLine(sStr);
                        m_ConnectionArray[iConnection].sAlias = sAlias;
                        SendPlayerList();
                    }
			        break;
		        case MessageType.CHOOSE_OPPONENT:
			        if(CreateGame(iConnection, Message) == false)
			        {
                        ilength = ByteArrayTo_ASCII_String(Message.cAlias, ref sAlias);
                        if (0 < ilength)
                        {
                            sAlias = TruncateString(sAlias, ilength);
                            sStr = string.Format("ChooseOpponent '{0}' FAILED", sAlias);
                            WriteStatusLine(sStr);
                        }
			        }
			        break;
		        case MessageType.LOCAL_CELL_SELECTION:
			        SendOpponentCellSelectionMessage(iConnection, Message.iCell);
			        SetCellInGame(iConnection, Message.iCell);
			        if(!CheckForWin(iConnection))
			        {
				        EndOfGame(iConnection);	// Check for end of game (all cells full)
			        }
			        break;
		        case MessageType.CLIENT_DISCONNECT:
			        sStr = string.Format("Client {0} Disconnected", m_ConnectionArray[iConnection].sAlias);
			        WriteStatusLine(sStr);
			        CloseGameAndConnection(iConnection);
			        break;
		    }
        }

        public Message_PDU DeSerialize_ReceivedMessage(byte[] bRecdByteArray)
        {   // Converts a byte array (from Receive) to a MyDetailsMessage structure (Fields have fixed length, but non-fixed data length)
            Message_PDU result_Message = Message_PDU_Initialise();
            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            result_Message.eMessageType = (MessageType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);

            iFieldOffset += m_iFieldSize1;
            result_Message.cAlias = new byte[MAX_ALIAS_LEN + 1];
            for (int iIndex = 0; iIndex < m_iFieldSize2; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cAlias[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += m_iFieldSize2;
            result_Message.cPlayerList = new byte[(MAX_ALIAS_LEN + 1) * 10];
            for (int iIndex = 0; iIndex < m_iFieldSize3; iIndex++)
            {   // Byte-by-byte copy data from the byte array into the data field  
                {
                    result_Message.cPlayerList[iIndex] = bRecdByteArray[iIndex + iFieldOffset];
                }
            }

            iFieldOffset += m_iFieldSize3;
            result_Message.iCell = ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize4;
            result_Message.eCode = (CodeType)ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize5;
            result_Message.iToken = ByteArrayToInt32(bRecdByteArray, iFieldOffset);
            iFieldOffset += m_iFieldSize6;
            result_Message.bFirstMove = BitConverter.ToBoolean(bRecdByteArray, iFieldOffset); // Converts a single byte within a byte array to a bool value   

            // Messages are natively aligned on 4-byte boundaries by C++ send(). To ensure compatibility, C# messages are adjusted in the
            // application code to truncate on 4-byte boundaries. The padding is added at the end of the message, so any bytes beyond the last
            // expected field can be ignored when receiving, i.e. no special decoding is needed for the padding when a message is DeSerializeed
            return result_Message;
        }

        bool CreateGame(int iConnection, Message_PDU RecdMessage)
        {
	        // Check initiating connection is not part of a game
	        if(IsConnectionPartOfActiveGame(iConnection) == true)
	        {
		        return false;
	        }
	        // Check chosen opponent alias is connected
	        string sAlias = "";
            int ilength = ByteArrayTo_ASCII_String(RecdMessage.cAlias, ref sAlias);
            if (0 >= ilength)
            {
                return false;
            }

            sAlias = TruncateString(sAlias, ilength);
            
            int iOpponentConnection = FindIndexOfConnectionForAlias(sAlias);
	        if(iOpponentConnection == -1)
	        {
		        return false;
	        }
	        // Check chosen opponent alias connection is not part of a game
	        if(IsConnectionPartOfActiveGame(iOpponentConnection) == true)
	        {
		        return false;
	        }
	        // Get next available game structure in array
	        int iGame = GetIndexOfUnusedGame();
	        if(iGame == -1)
	        {
		        return false;
	        }

	        m_GameArray[iGame].iConnection[0] = iConnection;
	        m_GameArray[iGame].iConnection[1] = iOpponentConnection;
	        m_GameArray[iGame].bInUse = true;
	        CountActiveGames();                 // Updates control
	        bool bFlag = CreateRandomBool();    // Determine who goes first
	
	        // Initialise game Grid
	        for(int j = 0; j < 3; j++)
	        {	
		        for(int k = 0; k < 3; k++)
		        {	
			        m_GameArray[iGame].iGrid[j,k] = -1; // -1 = cell unoccupied (0 = '0' 1 = 'X')
		        }	
	        }

            SendPlayerList(); // Send updated list of available players

	        // Send START_OF_GAME messages to both players
	        Message_PDU message = Message_PDU_Initialise(); 
	        message.eMessageType = MessageType.START_OF_GAME;
	        string sOpponentAlias = TruncateString(m_ConnectionArray[iOpponentConnection].sAlias, MAX_ALIAS_LEN);
            message.cAlias = StringTo_ASCII_ByteArray(sOpponentAlias);
	        message.iToken = GetPlayerToken(iGame, iConnection);	
	        message.bFirstMove = bFlag;
	        SendMessage(iConnection, message);

            sAlias = TruncateString(m_ConnectionArray[iConnection].sAlias, MAX_ALIAS_LEN);
            message.cAlias = StringTo_ASCII_ByteArray(sAlias);
  	        message.iToken = GetPlayerToken(iGame, iOpponentConnection);	
	        message.bFirstMove = !bFlag;
	        SendMessage(iOpponentConnection, message);

	        string sStr = string.Format("Game '{0}-{1}' created", 
		        m_ConnectionArray[m_GameArray[iGame].iConnection[0]].sAlias,
		        m_ConnectionArray[m_GameArray[iGame].iConnection[1]].sAlias);
	        WriteStatusLine(sStr);
	        return true;
        }

        int FindGameForConnection(int iConnection)			
        {
	        for(int i = 0; i < MAX_GAMES; i++)
	        {
                if (true == m_GameArray[i].bInUse && 
                (m_GameArray[i].iConnection[0] == iConnection ||
                m_GameArray[i].iConnection[1] == iConnection))
                {
                    return i;
                }
	        }
	        return -1;
        }

        int GetPlayerToken(int iGame, int iConnection)			
        {
	        // Lowest numbered connection will be given token '0'
	        if(iGame >= 0 && iGame < MAX_GAMES)
	        {
		        if(	(m_GameArray[iGame].iConnection[0] == iConnection && 
				        m_GameArray[iGame].iConnection[0] < m_GameArray[iGame].iConnection[1])
			        ||
			        (m_GameArray[iGame].iConnection[1] == iConnection && 
				        m_GameArray[iGame].iConnection[1] < m_GameArray[iGame].iConnection[0]) )
		        {
			        return 0;	// player identified by iConnection is '0'
		        }
		        else
		        {
			        return 1;	// player identified by iConnection is '1'
		        }
	        }
	        return -1;	// iGame out of range
        }

        void DoSetCellInGame(int iGame, int iCell, int iToken)			
        {
	        if((iToken == 0 || iToken == 1) && (iGame >= 0 && iGame < MAX_GAMES) 
		        && (iCell >= 0 && iCell < 9))
	        {
		        // Cell 0 [0][0], 1 [0][1], 2 [0][2]
		        //      3 [1][0], 4 [1][1], 5 [1][2]
		        //		6 [2][0], 7 [2][1], 8 [2][2]
		        m_GameArray[iGame].iGrid[iCell / 3, iCell % 3] = iToken;
	        }
	        else
	        {
		        string sStr = string.Format("Could not set cell {0}, in game {1}", iCell, iGame);
		        WriteStatusLine(sStr);
            }
        }

        void SetCellInGame(int iConnection, int iCell)			
        {
	        string sStr;
	        int iGame;
	        iGame  = FindGameForConnection(iConnection);
	        if(iGame  == -1)
	        {	
		        sStr = string.Format("Game for connection {0} not found", iConnection);
		        WriteStatusLine(sStr);
		        return;
	        }

	        // Check if player is '0' or 'X' ('0' is always assigned to the lowest numbered connection)
	        int iToken;
	        iToken = GetPlayerToken(iGame, iConnection);
	        if(iToken  == -1)
	        {	
		        sStr = string.Format("Token for connection {0} (game {1}) not found", iConnection, iGame);
		        WriteStatusLine(sStr);
		        return;
	        }
	        DoSetCellInGame(iGame, iCell, iToken);	
        }
	
        bool CheckForLine(int iGame, int iToken)
        {
	        // check for win on diagonals
	        if(m_GameArray[iGame].iGrid[1,1] == iToken &&
		        ( (m_GameArray[iGame].iGrid[0,0] == iToken &&
			        m_GameArray[iGame].iGrid[2,2] == iToken)
			        ||
		          (m_GameArray[iGame].iGrid[0,2] == iToken &&
			        m_GameArray[iGame].iGrid[2,0] == iToken) ) )
	        {
		        return true;
	        }
	        // check for horizontal line win
	        for(int j = 0; j < 3; j++)
	        {
		        if(m_GameArray[iGame].iGrid[j,0] == iToken &&
			        m_GameArray[iGame].iGrid[j,1] == iToken &&
			        m_GameArray[iGame].iGrid[j,2] == iToken)
		        {
			        return true;
		        }
	        }
	        // check for vertical line win
	        for(int j = 0; j < 3; j++)
	        {
		        if(m_GameArray[iGame].iGrid[0,j] == iToken &&
			        m_GameArray[iGame].iGrid[1,j] == iToken &&
			        m_GameArray[iGame].iGrid[2,j] == iToken)
		        {
			        return true;
		        }
	        }
	        return false;
        }

        bool CheckForWin(int iConnection)
        {
	        // Player iConnection has just moved
	        // Thus only that player could have won
	        int iGame;
	        iGame = FindGameForConnection(iConnection);
	        int iWinningToken;
	        iWinningToken = GetPlayerToken(iGame, iConnection);
	        // Seach for a line through the grid
	        bool bResult;
	        bResult = CheckForLine(iGame, iWinningToken);
	        if(bResult == true)
	        {
		        Message_PDU Message = Message_PDU_Initialise();
		        Message.eMessageType = MessageType.END_OF_GAME;
		        Message.eCode = CodeType.END_OF_GAME_WIN;
		        SendMessage(iConnection, Message);
		        int iOpponentConnection;
		        iOpponentConnection = GetConnectionOpponent(iConnection);
		        Message.eCode = CodeType.END_OF_GAME_LOSE;
		        SendMessage(iOpponentConnection, Message);
                RemoveGame(iGame);
	        }
	        return bResult;
        }

        bool EndOfGame(int iConnection)
        {
	        // If no cells free, signal end of game (Draw) to both players
	        int iGame;
	        iGame = FindGameForConnection(iConnection);
	
	        for(int j = 0; j < 3; j++)
	        {	
		        for(int k = 0; k < 3; k++)
		        {	
			        if(m_GameArray[iGame].iGrid[j,k] == -1) // -1 = cell unoccupied (0 = '0' 1 = 'X')
			        {
				        return false;	// At least one unoccupied cell
			        }
		        }	
	        }
	        Message_PDU Message = Message_PDU_Initialise();
            Message.eMessageType = MessageType.END_OF_GAME;
	        Message.eCode = CodeType.END_OF_GAME_DRAW;
	        SendMessage(iConnection, Message);
	        int iOpponentConnection;
	        iOpponentConnection = GetConnectionOpponent(iConnection);
	        SendMessage(iOpponentConnection, Message);
            RemoveGame(iGame);
	        return true;
        }

        void RemoveGame(int iGame)
        {
            m_GameArray[iGame].bInUse = false;
            CountActiveGames();                 // Updates control
            SendPlayerList();
        }
      
        void InformOpponent_GameEndedDueToOtherPlayerDisconnect(int iGame, int iDisconnectedClient_Connection)
        {
	        // First, identify which of the two connections in the game relates to the disconnected client, and thus which relates to the opponent
	        if(iDisconnectedClient_Connection == m_GameArray[iGame].iConnection[0])
	        {
		        CloseConnection(m_GameArray[iGame].iConnection[0]);
		        m_ConnectionArray[m_GameArray[iGame].iConnection[0]].bInUse = false;
	
		        // Send an END_OF_GAME (END_OF_GAME_CLIENT_DISCONNECT) message to the opponent
                Message_PDU Message = Message_PDU_Initialise();
                Message.eMessageType = MessageType.END_OF_GAME;
                Message.eCode = CodeType.END_OF_GAME_CLIENT_DISCONNECT;
                SendMessage(m_GameArray[iGame].iConnection[1], Message);

                m_GameArray[iGame].bInUse = false;
		        CountActiveGames(); // updates control
	        }
	        else if(iDisconnectedClient_Connection == m_GameArray[iGame].iConnection[1])
	        {
		        CloseConnection(m_GameArray[iGame].iConnection[1]);
		        m_ConnectionArray[m_GameArray[iGame].iConnection[1]].bInUse = false;
	
		        // Send an END_OF_GAME (END_OF_GAME_CLIENT_DISCONNECT) message to the opponent
                Message_PDU Message = Message_PDU_Initialise();
                Message.eMessageType = MessageType.END_OF_GAME;
                Message.eCode = CodeType.END_OF_GAME_CLIENT_DISCONNECT;
                SendMessage(m_GameArray[iGame].iConnection[0], Message);

		        m_GameArray[iGame].bInUse = false;
		        CountActiveGames(); // updates control
	        }
        }      
         
        void CloseGameAndConnection(int iConnection)
        {	
	        int iGame = FindGameForConnection(iConnection);
	        if(iGame != -1)
	        {
		        // Send a message informing the opponent that the game has ended due to the other player disconnecting
		        InformOpponent_GameEndedDueToOtherPlayerDisconnect(iGame, iConnection);
		        RemoveGame(iGame);
	        }
		    CloseConnection(iConnection);
        }

        void CloseConnection(int iConnection)
        {
            m_ConnectionArray[iConnection].bInUse = false;
            m_ConnectionArray[iConnection].sAlias = "";
            SendPlayerList();
            CountActiveConnections();
            try
            {
                m_ConnectionArray[iConnection].ConnectSock.Shutdown(SocketShutdown.Both);	// Close the connection
            }
            catch
            {   // Generic approach to dealing with situations such as socket not connected
            }
	        try
            {
                m_ConnectionArray[iConnection].ConnectSock.Close();			                // Close the socket
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }
        }

        void CloseAll()
        {
            StopTimer();
            for (int i = 0; i < MAX_CONNECTIONS; i++)	// Close active connections
	        {	
		        if(m_ConnectionArray[i].bInUse == true)
		        {
			        CloseConnection(i);
		        }
	        }
            try
            {
                m_ListenSocket.Close();		// Close the listen socket
            }
            catch
            {   // Generic approach to dealing with situations such as socket not created
            }

            Application.Exit();
        }
        
        private void ShutdownServer_Button_Click(object sender, EventArgs e)
        {
            CloseAll();
        }

        void WriteStatusLine(string sLine)
        {
            Events_ListBox.Items.Add(sLine);
        }

        bool CreateRandomBool()
        {
            int iRand = m_Random.Next(1,1000);
	        if(iRand >= 500)
	        {
                return true;
	        }
	        return false;
        }

        private void ClearEventList_Button_Click(object sender, EventArgs e)
        {
	        Events_ListBox.Items.Clear();  // Clear contents of Event ListBox
        }

        Message_PDU Message_PDU_Initialise()
        {
            Message_PDU M = new Message_PDU();
            M.eMessageType = MessageType.END_OF_GAME; // This field is always overwritten as used in every instance
            M.cAlias = StringTo_ASCII_ByteArray("");
            M.cPlayerList = StringTo_ASCII_ByteArray("");
            M.iCell = 0;
            M.eCode = CodeType.END_OF_GAME_DRAW;
            M.iToken = 0;
            M.bFirstMove = false;
            return M;
        }

        static byte[] StringTo_ASCII_ByteArray(string sStr)
        {   // Convert string into ASCII byte array, one char per byte           
            int iLength = sStr.Length;
            byte[] bArray = Encoding.ASCII.GetBytes(sStr); // Does not null-terminate the byte array
            byte[] bNullTerminatedArray = new byte[bArray.Length + 1];
            Buffer.BlockCopy(bArray, 0, bNullTerminatedArray, 0, bArray.Length);
            bNullTerminatedArray[bArray.Length] = 0; // Ensure null termination
            return bNullTerminatedArray;
        }

        static Int32 ByteArrayTo_ASCII_String(byte[] bArray, ref string sStr) // Returns length (to null terminator)
        {   // Convert byte array into ASCII string, one char per byte
            Int32 iASCIIz_length = 0; // Length to first zero (null termination)
            //Encoding.ASCII.GetBytes(sStr);
            sStr = System.Text.Encoding.UTF8.GetString(bArray);
            Int32 iStoragelength = sStr.Length;
            for (int iIndex = 0; iIndex < iStoragelength; iIndex++)
            {
                if (0 == bArray[iIndex])
                {   // Check for single-byte Zero character '0'
                    iASCIIz_length = iIndex;
                    break;
                }
            }
            return iASCIIz_length;
        }
        
        static byte[] Int32ToByteArray(Int32 iInt32)
        {   // The int parameter is treated as Int32 by GetBytes() so the resulting array contains 4 bytes
            byte[] bArray = BitConverter.GetBytes(iInt32);
            if (BitConverter.IsLittleEndian)
            {   // Make it platform independent
                Array.Reverse(bArray);
            }
            return bArray;
        }

        static Int32 ByteArrayToInt32(byte[] bArray, int iIndex)
        {   // Int32 contains 4 bytes. ToInt32() converts 4 bytes starting from iIndex to Int32 
            return (Int32)BitConverter.ToInt32(bArray, iIndex);
        }

        string TruncateString(string sStr, int iMaxLength)
        {
            if(sStr.Length > iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.Substring(0, iMaxLength);
            }
            return sStr;
        }

        void InitialiseServerAdvertiseAddressStructAndPort()
        {	// Initialise the IP address and Port number used for broadcasting the Server Advertisement messages
	        m_SendServerAdvertisementsEndPoint = new IPEndPoint(IPAddress.Broadcast, SERVER_ADVERTISE_PORT);
        }

        void Initialise_ServerAdvertisement()
        {
	        // Create m_SendServerAdvertisementsSocket and set it NonBlocking and Broadcast
            try
            {
                m_SendServerAdvertisementsSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                m_SendServerAdvertisementsSocket.Blocking = false; // Non-blocking mode
                m_SendServerAdvertisementsSocket.EnableBroadcast = true; // Set broadcast mode on
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                MessageBox.Show("CreateSocket() failed", "'Noughts and Crosses' Server");
                CloseAll();
            }
		
	        InitialiseServerAdvertiseAddressStructAndPort();
            Start_SendServerAdvertisement_Timer();
        }

        void Send_ServerAdvertisement_Broadcast()
        {	
	        ServerAdvertise_Message_PDU Message = new ServerAdvertise_Message_PDU();
	        Message.eMessageType =  MessageType.SERVER_ADVERTISE;
	        string sServerName = TruncateString("Noughts_and_Crosses_SERVER", MAX_SERVER_NAME_LEN);
            Message.cServerName = StringTo_ASCII_ByteArray(sServerName);
	        byte[] bLocalAddress = m_LocalEndPoint.Address.GetAddressBytes();
            Message.iIP_Address_Byte_1 = bLocalAddress[0];
            Message.iIP_Address_Byte_2 = bLocalAddress[1];
            Message.iIP_Address_Byte_3 = bLocalAddress[2];
            Message.iIP_Address_Byte_4 = bLocalAddress[3];
	        Message.iPort = m_iPort;

            byte[] bBuffer = Serialize_ServerAdvertise_Message_PDU(Message);
            int iBytesSent;
            try
            {
                iBytesSent = m_SendServerAdvertisementsSocket.SendTo(bBuffer, m_SendServerAdvertisementsEndPoint);
                if (0 < iBytesSent)
                {
                }
            }
            catch (SocketException ex)
            {
 		        MessageBox.Show("Server Advertise broadcast send failed","CNoughtsCrossesSERVER");
            }
	    }

       private byte[] Serialize_ServerAdvertise_Message_PDU(ServerAdvertise_Message_PDU Message)
        {   // Convert an Message_PDU structure into a byte array
            byte[] bByteArray = new byte[m_iSA_MessageSize];        // Create a byte array to hold the serialised form of the message PDU

            int iFieldOffset = 0; // Create an offset value which ensures the byte array is constructed with correct field positions and separations 
            byte[] bWorkingArray = BitConverter.GetBytes((Int32)Message.eMessageType);
            Buffer.BlockCopy(bWorkingArray/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, m_iSA_FieldSize1/*count*/);

            iFieldOffset += m_iSA_FieldSize1;                  // Adjust offset for second field
            int iDataLength2 = Message.cServerName.Length;  // Get the length of data actually stored in field2 (variable length data, fixed size field)
            Buffer.BlockCopy(Message.cServerName/*src*/, 0 /*source offset*/, bByteArray/*dest*/, iFieldOffset/*dest offset*/, iDataLength2/*count*/);

            iFieldOffset += m_iSA_FieldSize2;                  // Adjust offset for third field
            bByteArray[iFieldOffset] = Message.iIP_Address_Byte_1;
            iFieldOffset += m_iSA_FieldSize3;                  // Adjust offset for fourth field
            bByteArray[iFieldOffset] = Message.iIP_Address_Byte_2;
            iFieldOffset += m_iSA_FieldSize4;                  // Adjust offset for fifth field
            bByteArray[iFieldOffset] = Message.iIP_Address_Byte_3;
            iFieldOffset += m_iSA_FieldSize5;                  // Adjust offset for sixth field
            bByteArray[iFieldOffset] = Message.iIP_Address_Byte_4;
            
            iFieldOffset += m_iSA_FieldSize6;                  // Adjust offset for seventh field

            // The serialisation has been designed to be compatible with the C++ version of the application. The C++ serialisation is based on a
            // direct flattening of a structure containing fixed-size fields. The four individual byte values that hold the address are not aligned
            // on 4-byte boundaries, but the following port number IS. Therefore, a byte of padding is inserted in the message (at the sender side) 
            // to push the port field onto the next 4-byte boundary.
            iFieldOffset++; // Skip one byte to start port encoding on 4-byte boundary

            // Serialisation of port number
            // The message format and serialisation used in the C# version has been designed to interoperate with the pre-existing C++ version 
            // The port number is sent from the server as a 4-byte integer value of the format: high-byte-of-16-bit-value low-byte-of-16-bit-value 0 0  
            // the values in each byte are decimal values, for example the port number (decimal) 8000 is encoded, and received by clients, as:
            // 31 64 0 0 which tranlates into 31*256 + 64 = 8000. However, when converted directly as an Int16 we get 64*256 + 31 = 16415
            bByteArray[iFieldOffset++] = (byte) (Message.iPort / 256); // Create the MSB byte of the port value
            bByteArray[iFieldOffset++] = (byte) (Message.iPort % 256); // Create the LSB byte of the port value
            bByteArray[iFieldOffset++] = 0; // Complete the encoding of the additional bytes (so that the natively Int16 port value is encoded as Int32)
            bByteArray[iFieldOffset++] = 0; // Complete the encoding of the additional bytes (so that the natively Int16 port value is encoded as Int32)
 
            // Encoding the padding (to ensure message size is aligned with a 4-byte boundary): The bByteArray has been created to the full message
            // size including the pad. This byte array is automatically initialised such that all bytes contain zeros, therefore no additional
            // encoding is necessary to include the padding at the end of the message. The padding is always between 0 and 3 bytes.
            return bByteArray;
        }
    }
}
